import sys

from .chalk_factory import ChalkFactory
from .supports_color import detect_color_support


def create_chalk() -> ChalkFactory:
    mode = detect_color_support(sys.stdout)
    return ChalkFactory(mode)


chalk = create_chalk()



import re

from typing import List, Union

from .types import ColorMode, Code
from .ansi import (
    Mod,
    Color,
    BgColor,
    Ansi16Code,
    wrap_ansi_16,
)
from .utils import get_code_from_rgb, hex_to_rgb


class ChalkBuilder:
    def __init__(self, mode: ColorMode, codes: List[Code]):
        self._mode = mode
        self._codes = codes
        self._only_visible_if_supported = False

    def __call__(self, *args: object, sep: str = " ") -> str:
        if self._only_visible_if_supported and self._mode == ColorMode.AllOff:
            return ""

        if len(args) == 1 and isinstance(args[0], str):
            s = args[0]
        else:
            s = sep.join([str(arg) for arg in args])

        all_on = "".join([code.on for code in self._codes])
        all_off = "".join([code.off for code in self._codes])

        if "\u001b" in s:
            for code in self._codes:
                # Note: It seems to be necessary to replace the off code not just by
                # the on code, but keep the off code as well due to issues discussed here:
                # https://github.com/chalk/chalk/pull/335
                # https://github.com/chalk/chalk/issues/334
                # https://github.com/doowb/ansi-colors#nested-styling-bug
                s = s.replace(code.off, code.off + code.on)

        # Handle nested resets
        s = re.sub(
            "(\x1b\\[0m.*?\x1b\\[0m)",
            all_off + "\\1" + all_on,
            s,
        )

        # Handle newlines by closing and re-opening all styles
        if "\n" in s:
            s = re.sub(
                r"(\r?\n)",
                all_off + r"\1" + all_on,
                s,
            )

        return all_on + s + all_off

    # General style function

    def style(self, code: Union[Code, Ansi16Code]) -> "ChalkBuilder":
        if self._mode != ColorMode.AllOff:
            if isinstance(code, Ansi16Code):
                code = Code(on=wrap_ansi_16(code.on), off=wrap_ansi_16(code.off))
            self._codes.append(code)
        return self

    # Modifiers

    @property
    def reset(self) -> "ChalkBuilder":
        self.style(Mod.reset)
        return self

    @property
    def bold(self) -> "ChalkBuilder":
        self.style(Mod.bold)
        return self

    @property
    def dim(self) -> "ChalkBuilder":
        self.style(Mod.dim)
        return self

    @property
    def italic(self) -> "ChalkBuilder":
        self.style(Mod.italic)
        return self

    @property
    def underline(self) -> "ChalkBuilder":
        self.style(Mod.underline)
        return self

    @property
    def overline(self) -> "ChalkBuilder":
        self.style(Mod.overline)
        return self

    @property
    def inverse(self) -> "ChalkBuilder":
        self.style(Mod.inverse)
        return self

    @property
    def hidden(self) -> "ChalkBuilder":
        self.style(Mod.hidden)
        return self

    @property
    def strikethrough(self) -> "ChalkBuilder":
        self.style(Mod.strikethrough)
        return self

    @property
    def visible(self) -> "ChalkBuilder":
        self._only_visible_if_supported = True
        return self

    # Foreground colors

    @property
    def black(self) -> "ChalkBuilder":
        self.style(Color.black)
        return self

    @property
    def red(self) -> "ChalkBuilder":
        self.style(Color.red)
        return self

    @property
    def green(self) -> "ChalkBuilder":
        self.style(Color.green)
        return self

    @property
    def yellow(self) -> "ChalkBuilder":
        self.style(Color.yellow)
        return self

    @property
    def blue(self) -> "ChalkBuilder":
        self.style(Color.blue)
        return self

    @property
    def magenta(self) -> "ChalkBuilder":
        self.style(Color.magenta)
        return self

    @property
    def cyan(self) -> "ChalkBuilder":
        self.style(Color.cyan)
        return self

    @property
    def white(self) -> "ChalkBuilder":
        self.style(Color.white)
        return self

    @property
    def black_bright(self) -> "ChalkBuilder":
        self.style(Color.black_bright)
        return self

    @property
    def red_bright(self) -> "ChalkBuilder":
        self.style(Color.red_bright)
        return self

    @property
    def green_bright(self) -> "ChalkBuilder":
        self.style(Color.green_bright)
        return self

    @property
    def yellow_bright(self) -> "ChalkBuilder":
        self.style(Color.yellow_bright)
        return self

    @property
    def blue_bright(self) -> "ChalkBuilder":
        self.style(Color.blue_bright)
        return self

    @property
    def magenta_bright(self) -> "ChalkBuilder":
        self.style(Color.magenta_bright)
        return self

    @property
    def cyan_bright(self) -> "ChalkBuilder":
        self.style(Color.cyan_bright)
        return self

    @property
    def white_bright(self) -> "ChalkBuilder":
        self.style(Color.white_bright)
        return self

    @property
    def gray(self) -> "ChalkBuilder":
        self.style(Color.gray)
        return self

    @property
    def grey(self) -> "ChalkBuilder":
        self.style(Color.grey)
        return self

    # Background colors

    @property
    def bg_black(self) -> "ChalkBuilder":
        self.style(BgColor.black)
        return self

    @property
    def bg_red(self) -> "ChalkBuilder":
        self.style(BgColor.red)
        return self

    @property
    def bg_green(self) -> "ChalkBuilder":
        self.style(BgColor.green)
        return self

    @property
    def bg_yellow(self) -> "ChalkBuilder":
        self.style(BgColor.yellow)
        return self

    @property
    def bg_blue(self) -> "ChalkBuilder":
        self.style(BgColor.blue)
        return self

    @property
    def bg_magenta(self) -> "ChalkBuilder":
        self.style(BgColor.magenta)
        return self

    @property
    def bg_cyan(self) -> "ChalkBuilder":
        self.style(BgColor.cyan)
        return self

    @property
    def bg_white(self) -> "ChalkBuilder":
        self.style(BgColor.white)
        return self

    @property
    def bg_black_bright(self) -> "ChalkBuilder":
        self.style(BgColor.black_bright)
        return self

    @property
    def bg_red_bright(self) -> "ChalkBuilder":
        self.style(BgColor.red_bright)
        return self

    @property
    def bg_green_bright(self) -> "ChalkBuilder":
        self.style(BgColor.green_bright)
        return self

    @property
    def bg_yellow_bright(self) -> "ChalkBuilder":
        self.style(BgColor.yellow_bright)
        return self

    @property
    def bg_blue_bright(self) -> "ChalkBuilder":
        self.style(BgColor.blue_bright)
        return self

    @property
    def bg_magenta_bright(self) -> "ChalkBuilder":
        self.style(BgColor.magenta_bright)
        return self

    @property
    def bg_cyan_bright(self) -> "ChalkBuilder":
        self.style(BgColor.cyan_bright)
        return self

    @property
    def bg_white_bright(self) -> "ChalkBuilder":
        self.style(BgColor.white_bright)
        return self

    @property
    def bg_gray(self) -> "ChalkBuilder":
        self.style(BgColor.gray)
        return self

    @property
    def bg_grey(self) -> "ChalkBuilder":
        self.style(BgColor.grey)
        return self

    # rgb/hex

    def rgb(self, r: int, g: int, b: int) -> "ChalkBuilder":
        code = get_code_from_rgb(r, g, b, self._mode, background=False)
        if code is not None:
            self._codes.append(code)
        return self

    def hex(self, hex: str) -> "ChalkBuilder":
        r, g, b = hex_to_rgb(hex)
        return self.rgb(r, g, b)

    def bg_rgb(self, r: int, g: int, b: int) -> "ChalkBuilder":
        code = get_code_from_rgb(r, g, b, self._mode, background=True)
        if code is not None:
            self._codes.append(code)
        return self

    def bg_hex(self, hex: str) -> "ChalkBuilder":
        r, g, b = hex_to_rgb(hex)
        return self.bg_rgb(r, g, b)



from .chalk_factory import ChalkFactory
from .chalk_instance import chalk, create_chalk
from .types import ColorMode

VERSION = "0.1.8"

__all__ = [
    "chalk",
    "create_chalk",
    "ChalkFactory",
    "ColorMode",
]



from enum import Enum
from typing import NamedTuple


class ColorMode(Enum):
    AllOff = 0
    Basic16 = 1
    Extended256 = 2
    FullTrueColor = 3


class Code(NamedTuple):
    on: str
    off: str



"""
Inspired by:
https://github.com/chalk/supports-color/blob/main/index.js

With test cases:
https://github.com/chalk/supports-color/blob/main/test.js
"""

import os
import platform
import re
import sys
from typing import Optional, TextIO

from .types import ColorMode


def _get_env_force_color() -> Optional[ColorMode]:
    force_color = os.environ.get("FORCE_COLOR")

    if force_color is None:
        return None

    if force_color.lower() == "true" or len(force_color) == 0:
        return ColorMode.Basic16

    if force_color.lower() == "false":
        return ColorMode.AllOff

    try:
        level = int(force_color)
        return ColorMode(level)
    except ValueError:
        return None


def detect_color_support(stream: TextIO = sys.stdout) -> ColorMode:

    force_color = _get_env_force_color()

    is_tty = stream.isatty()

    # As a minor deviation from JS Chalk, we use the force color directly if given.
    # It feels like going into auto-detection mode even after a FORCE_COLOR value
    # is specified is not very intuitive (at least to me).
    if force_color is not None:
        return force_color

    elif os.environ.get("TERM") == "dumb":
        return ColorMode.AllOff

    elif platform.system() == "Windows":
        # https://stackoverflow.com/a/66554956/1804173
        windows_version = platform.version().split(".")
        if len(windows_version) == 3:
            try:
                major = int(windows_version[0])
                build = int(windows_version[2])
                if major >= 10:
                    if build >= 14931:
                        return ColorMode.FullTrueColor
                    elif build >= 10586:
                        return ColorMode.Extended256
            except ValueError:
                pass

        return ColorMode.Basic16

    elif "CI" in os.environ:
        if (
            any(
                name in os.environ
                for name in [
                    "TRAVIS",
                    "CIRCLECI",
                    "APPVEYOR",
                    "GITLAB_CI",
                    "GITHUB_ACTIONS",
                    "BUILDKITE",
                    "DRONE",
                ]
            )
            or os.environ.get("CI_NAME") == "codeship"
        ):
            return ColorMode.Basic16
        else:
            return ColorMode.AllOff

    elif "TEAMCITY_VERSION" in os.environ:
        team_city_version = os.environ["TEAMCITY_VERSION"]
        m = re.search(r"^(9\.(0*[1-9]\d*)\.|\d{2,}\.)", team_city_version)
        if m is not None:
            return ColorMode.Basic16
        else:
            return ColorMode.AllOff

    elif not is_tty:
        return ColorMode.AllOff

    elif os.environ.get("COLORTERM") == "truecolor":
        return ColorMode.FullTrueColor

    elif "TERM_PROGRAM" in os.environ:
        try:
            term_program_version_str = os.environ.get("TERM_PROGRAM_VERSION")
            if term_program_version_str is not None:
                term_program_version: Optional[int] = int(term_program_version_str.split(".")[0])
            else:
                term_program_version = None
        except ValueError:
            return ColorMode.Basic16
        term_program = os.environ.get("TERM_PROGRAM")
        if term_program == "iTerm.app":
            if term_program_version is not None:
                if term_program_version >= 3:
                    return ColorMode.FullTrueColor
                else:
                    return ColorMode.Extended256
            else:
                return ColorMode.Basic16
        elif term_program == "Apple_Terminal":
            return ColorMode.Extended256
        else:
            return ColorMode.Basic16

    elif "TERM" in os.environ and re.search(r"-256(color)?$", os.environ["TERM"]):
        return ColorMode.Extended256

    elif "TERM" in os.environ and re.search(
        r"^screen|^xterm|^vt100|^vt220|^rxvt|color|ansi|cygwin|linux", os.environ["TERM"]
    ):
        return ColorMode.Basic16

    elif "COLORTERM" in os.environ:
        return ColorMode.Basic16

    else:
        return ColorMode.AllOff



#!/usr/bin/env python

import setuptools  # type: ignore

if __name__ == "__main__":
    setuptools.setup(
        package_data={
            "yachalk": ["py.typed"],
        },
        include_package_data=True,
    )



from typing import Union

from .types import ColorMode, Code
from .ansi import (
    Mod,
    Color,
    BgColor,
    Ansi16Code,
    wrap_ansi_16,
)
from .utils import get_code_from_rgb, hex_to_rgb
from .chalk_builder import ChalkBuilder


class ChalkFactory:
    def __init__(self, mode: ColorMode = ColorMode.FullTrueColor):
        self._mode = mode

    # Internal helpers

    def _create_generator_from_ansi_16_code(self, ansi_16_code: Ansi16Code) -> "ChalkBuilder":
        if self._mode != ColorMode.AllOff:
            on, off = ansi_16_code
            codes = [Code(on=wrap_ansi_16(on), off=wrap_ansi_16(off))]
        else:
            codes = []
        return ChalkBuilder(
            mode=self._mode,
            codes=codes,
        )

    # Color mode control

    def get_color_mode(self) -> ColorMode:
        return self._mode

    def set_color_mode(self, mode: ColorMode) -> None:
        self._mode = mode

    def disable_all_ansi(self) -> None:
        self.set_color_mode(ColorMode.AllOff)

    def enable_basic_colors(self) -> None:
        self.set_color_mode(ColorMode.Basic16)

    def enable_extended_colors(self) -> None:
        self.set_color_mode(ColorMode.Extended256)

    def enable_full_colors(self) -> None:
        self.set_color_mode(ColorMode.FullTrueColor)

    # General style function

    def style(self, code: Union[Code, Ansi16Code]) -> ChalkBuilder:
        g = ChalkBuilder(self._mode, [])
        g.style(code)
        return g

    # Modifiers

    @property
    def reset(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.reset)

    @property
    def bold(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.bold)

    @property
    def dim(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.dim)

    @property
    def italic(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.italic)

    @property
    def underline(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.underline)

    @property
    def overline(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.overline)

    @property
    def inverse(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.inverse)

    @property
    def hidden(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.hidden)

    @property
    def strikethrough(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Mod.strikethrough)

    @property
    def visible(self) -> "ChalkBuilder":
        return ChalkBuilder(mode=self._mode, codes=[]).visible

    # Foreground colors

    @property
    def black(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.black)

    @property
    def red(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.red)

    @property
    def green(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.green)

    @property
    def yellow(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.yellow)

    @property
    def blue(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.blue)

    @property
    def magenta(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.magenta)

    @property
    def cyan(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.cyan)

    @property
    def white(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.white)

    @property
    def black_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.black_bright)

    @property
    def red_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.red_bright)

    @property
    def green_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.green_bright)

    @property
    def yellow_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.yellow_bright)

    @property
    def blue_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.blue_bright)

    @property
    def magenta_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.magenta_bright)

    @property
    def cyan_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.cyan_bright)

    @property
    def white_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.white_bright)

    @property
    def gray(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.gray)

    @property
    def grey(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(Color.grey)

    # Background colors

    @property
    def bg_black(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.black)

    @property
    def bg_red(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.red)

    @property
    def bg_green(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.green)

    @property
    def bg_yellow(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.yellow)

    @property
    def bg_blue(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.blue)

    @property
    def bg_magenta(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.magenta)

    @property
    def bg_cyan(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.cyan)

    @property
    def bg_white(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.white)

    @property
    def bg_black_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.black_bright)

    @property
    def bg_red_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.red_bright)

    @property
    def bg_green_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.green_bright)

    @property
    def bg_yellow_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.yellow_bright)

    @property
    def bg_blue_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.blue_bright)

    @property
    def bg_magenta_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.magenta_bright)

    @property
    def bg_cyan_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.cyan_bright)

    @property
    def bg_white_bright(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.white_bright)

    @property
    def bg_gray(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.gray)

    @property
    def bg_grey(self) -> ChalkBuilder:
        return self._create_generator_from_ansi_16_code(BgColor.grey)

    # rgb/hex

    def rgb(self, r: int, g: int, b: int) -> ChalkBuilder:
        code = get_code_from_rgb(r, g, b, self._mode, background=False)
        return ChalkBuilder(self._mode, [code] if code is not None else [])

    def hex(self, hex: str) -> ChalkBuilder:
        r, g, b = hex_to_rgb(hex)
        return self.rgb(r, g, b)

    def bg_rgb(self, r: int, g: int, b: int) -> ChalkBuilder:
        code = get_code_from_rgb(r, g, b, self._mode, background=True)
        return ChalkBuilder(self._mode, [code] if code is not None else [])

    def bg_hex(self, hex: str) -> ChalkBuilder:
        r, g, b = hex_to_rgb(hex)
        return self.bg_rgb(r, g, b)



import re

from typing import Optional, Tuple

from .types import ColorMode, Code
from . import ansi


def hex_to_rgb(hex: str) -> Tuple[int, int, int]:

    if hex.startswith("#"):
        hex = hex[1:]

    if len(hex) == 3:
        hex = hex[0] + hex[0] + hex[1] + hex[1] + hex[2] + hex[2]

    hex = hex.lower()
    m = re.match(r"^([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$", hex)

    if m is None:
        raise ValueError(f"'{hex}' is not a valid hex literal")

    return int(m.group(1), 16), int(m.group(2), 16), int(m.group(3), 16)


def get_code_from_rgb(r: int, g: int, b: int, mode: ColorMode, background: bool) -> Optional[Code]:
    if mode == ColorMode.FullTrueColor:
        code = ansi.wrap_ansi_16m(r, g, b, background=background)
    elif mode == ColorMode.Extended256:
        code_256 = ansi.rgb_to_ansi_256(r, g, b)
        code = ansi.wrap_ansi_256(code_256, background=background)
    elif mode == ColorMode.Basic16:
        code_256 = ansi.rgb_to_ansi_256(r, g, b)
        code_16 = ansi.ansi_256_to_ansi_16(code_256)
        code = ansi.wrap_ansi_16(code_16, background=background)
    else:
        return None

    close = ansi.BG_COLOR_CLOSE if background else ansi.COLOR_CLOSE
    return Code(on=code, off=ansi.wrap_ansi_16(close))



"""
Inspired by:
https://github.com/chalk/ansi-styles/blob/main/index.js
"""

import math

from typing import NamedTuple


class Ansi16Code(NamedTuple):
    on: int
    off: int


class Mod:
    reset = Ansi16Code(0, 0)
    bold = Ansi16Code(1, 22)  # 21 isn't widely supported and 22 does the same thing
    dim = Ansi16Code(2, 22)
    italic = Ansi16Code(3, 23)
    underline = Ansi16Code(4, 24)
    overline = Ansi16Code(53, 55)
    inverse = Ansi16Code(7, 27)
    hidden = Ansi16Code(8, 28)
    strikethrough = Ansi16Code(9, 29)


COLOR_CLOSE = 39
BG_COLOR_CLOSE = 49


class Color:
    black = Ansi16Code(30, COLOR_CLOSE)
    red = Ansi16Code(31, COLOR_CLOSE)
    green = Ansi16Code(32, COLOR_CLOSE)
    yellow = Ansi16Code(33, COLOR_CLOSE)
    blue = Ansi16Code(34, COLOR_CLOSE)
    magenta = Ansi16Code(35, COLOR_CLOSE)
    cyan = Ansi16Code(36, COLOR_CLOSE)
    white = Ansi16Code(37, COLOR_CLOSE)

    black_bright = Ansi16Code(90, COLOR_CLOSE)
    red_bright = Ansi16Code(91, COLOR_CLOSE)
    green_bright = Ansi16Code(92, COLOR_CLOSE)
    yellow_bright = Ansi16Code(93, COLOR_CLOSE)
    blue_bright = Ansi16Code(94, COLOR_CLOSE)
    magenta_bright = Ansi16Code(95, COLOR_CLOSE)
    cyan_bright = Ansi16Code(96, COLOR_CLOSE)
    white_bright = Ansi16Code(97, COLOR_CLOSE)

    gray = Ansi16Code(90, COLOR_CLOSE)
    grey = Ansi16Code(90, COLOR_CLOSE)


class BgColor:
    black = Ansi16Code(40, BG_COLOR_CLOSE)
    red = Ansi16Code(41, BG_COLOR_CLOSE)
    green = Ansi16Code(42, BG_COLOR_CLOSE)
    yellow = Ansi16Code(43, BG_COLOR_CLOSE)
    blue = Ansi16Code(44, BG_COLOR_CLOSE)
    magenta = Ansi16Code(45, BG_COLOR_CLOSE)
    cyan = Ansi16Code(46, BG_COLOR_CLOSE)
    white = Ansi16Code(47, BG_COLOR_CLOSE)

    black_bright = Ansi16Code(100, BG_COLOR_CLOSE)
    red_bright = Ansi16Code(101, BG_COLOR_CLOSE)
    green_bright = Ansi16Code(102, BG_COLOR_CLOSE)
    yellow_bright = Ansi16Code(103, BG_COLOR_CLOSE)
    blue_bright = Ansi16Code(104, BG_COLOR_CLOSE)
    magenta_bright = Ansi16Code(105, BG_COLOR_CLOSE)
    cyan_bright = Ansi16Code(106, BG_COLOR_CLOSE)
    white_bright = Ansi16Code(107, BG_COLOR_CLOSE)

    gray = Ansi16Code(100, BG_COLOR_CLOSE)
    grey = Ansi16Code(100, BG_COLOR_CLOSE)


_ANSI_BACKGROUND_OFFSET = 10


def wrap_ansi_16(code: int, background: bool = False) -> str:
    offset = _ANSI_BACKGROUND_OFFSET if background else 0
    return f"\x1b[{code + offset}m"


def wrap_ansi_256(code: int, background: bool = False) -> str:
    offset = _ANSI_BACKGROUND_OFFSET if background else 0
    return f"\x1b[{38 + offset};5;{code}m"


def wrap_ansi_16m(r: int, g: int, b: int, background: bool = False) -> str:
    offset = _ANSI_BACKGROUND_OFFSET if background else 0
    return f"\x1b[{38 + offset};2;{r};{g};{b}m"


def rgb_to_ansi_256(r: int, g: int, b: int) -> int:
    if r == g and g == b:
        if r < 8:
            return 16
        if r > 248:
            return 231
        return round(((r - 8) / 247) * 24) + 232

    return 16 + (36 * round(r / 255 * 5)) + (6 * round(g / 255 * 5)) + round(b / 255 * 5)


def ansi_256_to_ansi_16(code: int) -> int:
    if code < 8:
        return 30 + code

    if code < 16:
        return 90 + (code - 8)

    if code >= 232:
        r = (((code - 232) * 10) + 8) / 255
        g = r
        b = r
    else:
        code -= 16
        remainder = code % 36

        r = math.floor(code / 36) / 5
        g = math.floor(remainder / 6) / 5
        b = (remainder % 6) / 5

    value = max(r, g, b) * 2

    if value == 0:
        return 30

    result = 30 + ((round(b) << 2) | (round(g) << 1) | round(r))

    if value == 2:
        result += 60

    return result
